function [x_sol, x_itr, f_itr]=solver(func,x_guess,varargin)

%%
% ********************************************************************
% Course Homework
%               2.29: Numerical Fluid Mechanincs, Spring 2007
% Problem Set #2, Problem #1: MATLAB Workshop Problem
% 03/08/07
% ********************************************************************
% Reza Karimi:
%                            Graduate Student
%                    Mechanical Engineering Department
%                  Massachusetts Institute of Technology
% Email:  rkarimi@mit.edu
% Office: NE47-319, 500 Technology Square, Cambridge, MA 02139
% Phone: (617) 324-7605
% ********************************************************************
%
%% How to call
%  [x_solution, x_iterations, f_iterations] = solver(func_name)
%  [x_solution, x_iterations, f_iterations] = solver(func_name, x_guess)
%  [x_solution, x_iterations, f_iterations] = solver(func_name, x_guess, OPTIONS)
%  
%  % Comments
%  x_solution scalar final solution
%  x_iteration\f_iterations matrix: n*2 for Newton Method or n*3 for for other methods
%       Columns in Bracketing Methods     x_lower x_upper x_new 
%       Columns in Secant                 x_0     x_1     x_new (note that x_new=x_2, x_old=x_1)
%       Columns in Newton                 x_old   x_new    
%
%% Inputs\Options\Setting
% The input function (to be solved) should be either a function handle or
% a string (like ?3*x^3-5*x+1). This is the minimum input. Options
% include:
% a)	Method: Newton, Secant, Bi-Section, False-Position, Modified False-Position
% b)	Initial guess: it can be two numbers for methods like Bi-Section
% c)	Derivative of f (note that you can compute the derivative if user provides you with a string as solution equation)
% d)	Absolute tolerance on x or f(right now abs_er_f is set to zero)
% e)	Relative error on x
% f)	Maximum number of iterations
% g)	Plot:    on\off 1\0
% h)	Display: on\off 1\0
%
%% Example Calls:
%  [x_f]=solver('x-sin(x)');
%  [x_f]=solver(@func);   % where func is a function handle
%
%  [x_f,x_it,f_itr]=solver(@func);
%  [x_f,x_it,f_itr]=solver(@func,2);
%
%  [x_f,x_it,f_itr]=solver(@func,[-2 2],'method','Bi-Section');
%  [x_f,x_it,f_itr]=solver(@func,[-2 2],'MEthod','bi-secTION');
%
%  [x_f,x_it,f_itr]=solver(@func,[-2 2],'plot','off');
%  [x_f,x_it,f_itr]=solver(@func,[-2 2],'method','Bi-Section','plot','off');
%  [x_f,x_it,f_itr]=solver(@func,[-2 2],'plot','off','method','Bi-Section');
%
%  [x_f,x_it,f_itr]=solver('x-sin(x)',2,'method','Newton');
%  [x_f,x_it,f_itr]=solver(@func,2,'method','Newton','f_derivative',@d_func);
%
%  [x_f,x_it,f_itr]=solver('x-sin(x)',2,'max_iteration',10, 'method','Newton');
%  [x_f,x_it,f_itr]=solver(@func     ,2,'max_iteration',10,'abs_tolerance',1e-8,'rel_tolerance',1e-6,'plot','on');
%%

if nargin<1
    error('A minimum of one input function is required.')
elseif isa(func,'function_handle')
    do_handle_f=1;
    func_str=[];
elseif isstr(func)
    func=lower(func);
    func_str=[': ' func]; % for plot_title
    try
        x=0;
        fval=eval(func);
        do_handle_f=0;
    catch
        error('The string function have to be evaluated solely by "x" value.')
    end
else
    error('The fisrt input have to be either a function handle or a string.')
end

%% Default Setting
do_plot=1;
do_display=1; display_jump=1; %Internal Setting (not modified by function call)
method='bi-section';double_sign=0;
max_itr=20;
rel_tol=1e-12; % Used for x
rel_tol_f=0;   % Used for f, Practically not used by setting to 0. (Internal Setting)
abs_tol=1e-9;
d_func=[];

%% Input Processing
if nargin>2
    opt=(nargin-2)/2;
    if opt==round(opt)
        for i=1:opt
            if isstr(varargin{2*i-1})
                opt_v=varargin{2*i};
                switch lower(varargin{2*i-1})
                    case 'rel_tolerance'
                        if isnumeric(opt_v) && opt_v>0
                            rel_tol=opt_v;
                        else
                            error(['Unacceptable value for rel_tolerance option.'])
                        end
                    case 'abs_tolerance'
                        if isnumeric(opt_v) && opt_v>0
                            abs_tol=opt_v;
                        else
                            error(['Unacceptable value for abs_tolerance option.'])
                        end
                    case 'max_iteration'
                        if isnumeric(opt_v) && opt_v>0 && round(opt_v)==opt_v
                            max_itr=opt_v;
                        else
                            error(['Unacceptable value for max_iteration option.'])
                        end
                    case 'plot'
                        if strcmp(opt_v,'on') || (isnumeric(opt_v) && opt_v==1)
                            do_plot=1;
                        elseif strcmp(opt_v,'off') || (isnumeric(opt_v) && opt_v==0)
                            do_plot=0;
                        else
                            error(['Unacceptable value for plot option.'])
                        end
                      case 'display'
                        if strcmp(opt_v,'on') || (isnumeric(opt_v) && opt_v==1)
                            do_display=1;
                        elseif strcmp(opt_v,'off') || (isnumeric(opt_v) && opt_v==0)
                            do_display=0;
                        else
                            error(['Unacceptable value for display option.'])
                        end
                    case 'method'
                        if isstr(opt_v)
                            opt_v=lower(opt_v);
                            switch opt_v
                                case {'newton', 'secant', 'bi-section', 'false-position', 'modified false-position'}
                                    method=opt_v;
                                otherwise
                                    error(['Unacceptable value for method option.'])
                            end
                        else
                            error('The method name should be a string.')
                        end
                    case 'f_derivative'
                        if ~isa(opt_v,'function_handle')
                            error('Derivative of function is only acceptable as a function handle.')
                        else
                            d_func=opt_v;
                        end
                    otherwise
                        error(['"' lower(varargin{2*i-1}) '" is not an acceptable option name.'])
                end
            else
                error('The options name should be a string.')
            end

        end
    else
        error('The options are not paired.')
    end
end

switch method
    case 'newton'
        n_guess=1;
        if do_handle_f
            if isempty(d_func)
                error('Derivative of function is required as a function handle for Newton method applied to a function handle.')
            end
        else
            if ~isempty(d_func)
                error('The derivative of symbolic function will be computed by the program itself!')
            end
            d_func=diff(func,'x');
        end
    case {'secant', 'bi-section', 'false-position', 'modified false-position'}
        n_guess=2;
        double_sign=1;
        if ~isempty(d_func)
            %fprintf('\n')
            %fprintf('\tWarning: no derivative is required for these methods: Secant, Bi-Section, False-Position, Modified False-Position')
        end
        if strcmp(method,'secant')
            double_sign=0;
        end
end

if nargin>1 && ~isempty(x_guess)
    if isnumeric(x_guess)
        if prod(size(x_guess))~=n_guess
            error('Size of initial guess does not match the method')
        elseif n_guess==2 && x_guess(1)==x_guess(2)
            error('Two different initial guesses are required for these methods: Secant, Bi-Section, False-Position, Modified False-Position')
        end
    else
        error('Initial guess shoud be a number or empty value')
    end
else
    if n_guess==2
        x_guess=[-1 1];
    elseif n_guess==1
        x_guess=0;
    end
end

if ~do_handle_f % Transforming string function into handle functio
    eval(['func=@(x) ' func ';']);
    if ~isempty(d_func)
        d_func=@(x) subs(f,'x',x);
    end
end

if n_guess==2
    x_u=x_guess(2);f_u=func(x_u);
    x_l=x_guess(1);f_l=func(x_l);
    if abs(f_u)<abs(f_l) && double_sign % Order of initial guess matters for Secant's method
        x_o=x_u;
        f_o=f_u;
    else
        x_o=x_l;
        f_o=f_l;
    end
    if double_sign
        if (f_u*f_l)>0
            error(['Initial guess of x= ' num2str(x_u) ' ,' num2str(x_l) 'does not hold the sign variation prerequisite'])
        else
            if f_u<0
                x_u=x_l;f_u=f_l;
                x_l=x_guess(2);f_l=func(x_l);
            end
        end
    end
else
    x_o=x_guess;
    f_o=func(x_o);
    df_o=d_func(x_o);
end


if f_o==0
    x_sol=x_o;, x_itr, f_itr
    fprintf('\n\n \t x=%+20.16f is the exact solution.\n',x_o)
end

%% Main Part

x_rel_er=inf;
x_abs_er=inf;
f_rel_er=inf;
% abs_tol_f=abs_tol;
abs_tol_f=0;

if do_display
    n_frmt='%+10.8f'; %fprintf format
    er_frmt='%+5.5e';
    switch method
        case {'bi-section', 'false-position', 'modified false-position'}
            fprintf('\n\n k \t x_l \t\t x_u \t\t x_n \t\t f_l \t\t f_u \t\t f_n \t\t x_rel_error')
            fprintf('\n------------------------------------------------------------------------------------------------------------------------------%')
        case 'secant'
            fprintf('\n\n k \t x_0 \t\t x_1 \t\t x_2 \t\t f_0 \t\t f_1 \t\t f_2 \t\t df_x_1 \t x_rel_error')
            fprintf('\n-----------------------------------------------------------------------------------------------------------------------------------------%')
        case 'newton'
            fprintf('\n\n k \t x_o \t\t x_n \t\t f_o \t\t f_n  \t\t df_o \t\t x_rel_error%')
            fprintf('\n--------------------------------------------------------------------------------------------------------%')
    end
end
warning off
i_up_fixed=0;
i_lo_fixed=0;
k=0; %number of iterations

while (k<=max_itr && abs(f_o)>abs_tol_f && f_rel_er>rel_tol_f && x_abs_er>abs_tol && x_rel_er>rel_tol)

    switch method
        case 'bi-section'
            x_n=(x_u+x_l)/2;
        case 'false-position'
            m=(f_u-f_l)/(x_u-x_l);
            x_n=x_l-f_l/m;
        case 'modified false-position'
            if i_lo_fixed>1
                f_l=f_l/2;
            elseif i_up_fixed>1
                f_u=f_u/2;
            end
            m=(f_u-f_l)/(x_u-x_l);
            x_n=x_l-f_l/m;

        case 'secant'  % (xl,x_o=x_u)->x_n
            m=(f_u-f_l)/(x_u-x_l);
            x_n=x_l-f_l/m;
            df_o=m;
        case 'newton'
            df_o=d_func(x_o);
            x_n=x_o-f_o/d_func(x_o);
    end
    f_n=func(x_n);

    x_abs_er=abs(x_n-x_o);
    if x_n~=0
        x_rel_er=x_abs_er/abs(x_n);
    else
        x_rel_er=inf;
    end
    if f_n~=0
        f_rel_er=abs((f_n-f_o)/f_n);
    else
        f_rel_er=inf;
    end
    if do_display && rem(k,display_jump)==0
        switch method
            case {'bi-section','false-position','modified false-position'}
                fprintf(['\n %d \t ' n_frmt '\t ' n_frmt '\t ' n_frmt '\t ' er_frmt '\t ' er_frmt '\t ' er_frmt '\t ' er_frmt '%%' ],k,x_l,x_u,x_n,f_l,f_u,f_n,x_rel_er*100)
            case 'secant'
                fprintf(['\n %d \t ' n_frmt '\t ' n_frmt '\t ' n_frmt '\t ' er_frmt '\t ' er_frmt '\t ' er_frmt '\t ' er_frmt '\t ' er_frmt '%%' ],k,x_l,x_u,x_n,f_l,f_u,f_n,df_o,x_rel_er*100)
            case 'newton'
                fprintf(['\n %d \t ' n_frmt '\t ' n_frmt '\t ' er_frmt '\t ' er_frmt '\t ' er_frmt '\t ' er_frmt '%%' ],k,x_o,x_n,f_o,f_n,df_o,x_rel_er*100)
        end
    end
    k=k+1;
    if n_guess==2
        x_itr(k,:)=[x_l x_u x_n];
        f_itr(k,:)=[f_l f_u f_n];
        if double_sign
            if f_n>=0
                x_u=x_n;
                f_u=f_n;
                i_lo_fixed=i_lo_fixed+1;
                i_up_fixed=0;
            elseif f_n<0
                x_l=x_n;
                f_l=f_n;
                i_up_fixed=i_up_fixed+1;
                i_lo_fixed=0;
            end
        else
            x_l=x_u;
            f_l=f_u;
            x_u=x_n;
            f_u=f_n;
        end
    else
       x_itr(k,:)=[x_o x_n];f_itr(k,:)=[f_o f_n];
    end

    
    if f_n==0
        fprintf('\n\n \t x=%+20.16f is the exact solution.\n',x_n)
    end
    x_o=x_n;
    f_o=f_n;
end
x_sol=x_n;
warning on
fprintf('\n\n Final Solution: f(x=%+20.16f)=%+8.6e with %s method and %d iterations \n ',x_sol,func(x_sol),upper(method),k-1)

%% plot
if do_plot
    AX=plotyy([1:size(x_itr,1)],x_itr(:,end),[1:size(x_itr,1)],abs(x_itr(:,end)-x_itr(:,end-1))./abs(x_itr(:,end)),@plot,@semilogy); %

    hold on, box on, grid on
    FN_Size_A=16;
    FN_Size_T=14;
    set(AX(1),'fontsize',14),set(AX(2),'fontsize',14)
    xlabel('n (Iteration Number)','Interpreter','latex','fontsize',FN_Size_A,...
        'FontAngle','italic','FontWeight','bold','FontName','Times New Roman')
    ylabel(AX(1),'x','Interpreter','latex','fontsize',FN_Size_A,'FontAngle','italic','FontWeight','bold','FontName','Times New Roman')
        ylabel(AX(2),'$\epsilon_a$ (relative error of x)','Interpreter','latex','fontsize',FN_Size_A,'FontAngle','italic','FontWeight','bold','FontName','Times New Roman')

    title (['Root of Equation' func_str],'Interpreter','latex',...
        'fontsize',FN_Size_T,'FontAngle','italic','FontWeight','bold','FontName','Times New Roman')

    %legend('\sigma_{rr}','\sigma_{\theta\theta}','\sigma_{Mises}','Location','Southwest')
    set(gcf,'PaperType','usLetter','PaperOrientation','portrait','PaperUnits','Centimeters','PaperPosition',[1 1 20 18])
    set(gcf,'Position',[50 50 [20 18]*30],'color','w')
    %
end